; srbm.lisp - a semi-random beat maker. there are things to TWEAK

(require :asdf)
(asdf:load-system :smolmidi)    ; https://thrig.me/src/smolmidi.git
(asdf:load-system :statslite)   ; https://thrig.me/src/statslite.git

(block nil (setq *random-state* (make-random-state t)) (return))

; TWEAK these can be fiddled with, though changing measures may affect
; the longpat function
(defconstant +bpm+       16     "beats per measure")
(defconstant +beatdur+   24     "duration of a beat (MIDI dtime)")
(defconstant +measures+  64     "how many measures to generate")
(defconstant +tempo+     630000)

; TWEAK these are for a "Magic Techno Drums" SoundFont
(defconstant +pitch-min+ 24)
(defconstant +pitch-max+ 73)

(defmacro coinflip () `(zerop (random 2)))

(defun random-pitch (min max)
  (+ +pitch-min+ (random (1+ (- +pitch-max+ +pitch-min+)))))

(let ((seen (make-hash-table)))
  (defun random-pitch-unique (&aux pitch)
    (loop do
          (setf pitch (random-pitch +pitch-min+ +pitch-max+))
          unless (eq pitch (gethash pitch seen)) do (loop-finish))
    (setf (gethash pitch seen) pitch)))

; TWEAK different velocity levels can be handy
(defun random-velocity () (if (coinflip) 90 45))
(defun velonoise () (+ (random 4) (random 4) (random 4)))

(defun pattern-random (onsets total)
  "turn on N onsets in M total beats"
  (let ((pattern (make-array total
                             :element-type 'bit
                             :initial-element 0)))
    (loop for i from 0 while (plusp total) do
          (when (< (random 1.0) (/ onsets total))
            (setf (aref pattern i) 1)
            (decf onsets))
          (decf total))
    pattern))

(defun pattern-onset-deltas (pattern)
  "distances betwen onsets in a pattern"
  (let ((deltas) (len (array-dimension pattern 0)))
    (loop for i from 0 for n across pattern do
          (when (plusp n)
            (loop for j from (1+ i) do
                  (when (plusp (aref pattern (mod j len)))
                    (push (- j i) deltas)
                    (loop-finish)))))
    deltas))

(defun pattern-score-sdev (pattern)
  "lower score means a more regular beat"
  (let ((stats (statslite:statslite (pattern-onset-deltas pattern))))
    (statslite:statslite-sd stats)))

; too low a score here could be too regular
(defun pattern-semirandom (onsets beats trials)
  "more trials tend towards a more regular beat pattern"
  (let ((best most-positive-single-float) (pattern))
    (loop repeat trials do
          (let* ((p (pattern-random onsets beats))
                 (score (pattern-score-sdev p)))
            (when (and (> score 0.8) (< score best))
              (psetf best score pattern p))))
    pattern))

; the beat-count for 16 bpm should probably be 8 or 4 or 2, but this
; code will do something with other values
(defun beats (track beat-count
              &key (dur +beatdur+) (offset 0) (measures +measures+)
              (pitch 60) (velocity 90))
  "generate some number of beats per measure"
  (declare (fixnum beat-count))
  (when (> beat-count +bpm+) (error "beat-count cannot be that high"))
  (let ((maxbeat (* measures +bpm+))
        (beats (truncate (/ +bpm+ beat-count))))
    (loop while (plusp maxbeat) do
          (smolmidi:note-on track offset pitch (+ velocity (velonoise)))
          (smolmidi:note-off track dur pitch 0)
          (setf offset (* dur (1- beats)))
          (decf maxbeat beats))
    (values track offset)))

(defun four (&key (dur +beatdur+) (measures +measures+) (offset 0))
  "put down four beats per measure, possibly with an initial offset"
  (let ((track (smolmidi:new-track)))
    (smolmidi:tempo track 0 +tempo+)
    (beats track 4 :dur dur :measures measures :offset offset
           :pitch (random-pitch-unique)
           :velocity (random-velocity))))

; TWEAK the arguments to pattern-semirandom could be fiddled with
(defun longpat (&key (dur +beatdur+) (measures +measures+) (div 2))
  "a longer random pattern"
  (let* ((track (smolmidi:new-track))
         (total-beats (/ (* +bpm+ measures) div))
         (pattern (pattern-semirandom
                    (truncate (/ total-beats 3)) total-beats (* div 30)))
         (pitch (random-pitch-unique))
         (velo (random-velocity)))
    (smolmidi:patternize track pattern :dtime dur :times div
                         :pitch pitch :velocity velo)))

; TWEAK customize the tracks, what they do, etc
(defun make-tracks ()
  "make a list of MIDI tracks"
  (list
    (four)
    (four :offset (* +beatdur+ 2))
    (longpat :div 8)
    (longpat :div 16)
    (longpat :div 32)
    )) ; not kosher but it makes fiddling with the above lines easier in vi

(smolmidi:write-midi-file "out.midi" (make-tracks))